/** @file bitostream_inline.h
 * Bit-oriented output stream buffer.
 */

/*
 * begin        : Tue Nov 9 2001
 * copyright    : (C) 2001 by Ondrej Pecta, Stepan Vondrak
 */

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it according to instructions in massiv/doc/licenses/source.txt.       *
 *                                                                         *
 ***************************************************************************/

#ifndef MASSIV_BITOSTREAM_H
#include "net/bitostream.h"
#endif



namespace Massiv {
namespace Core {


BitOStream::BitOStream()
    {
    reset();
    }


BitOStream::BitOStream
    (
    int size
    )
    {
    reset( size );
    }


BitOStream::~BitOStream()
    {
    }


BitIOStreamBuffer BitOStream::unbind()
    {
    BitIOStreamBuffer buf;
    buf.reserve( DEFAULT_STRBUF_SIZE );
    return bind( buf, false );
    }


BitIOStreamBuffer BitOStream::reset
    (
    int size
    )
    {
    BitIOStreamBuffer buf;
    buf.reserve( size );
    return bind( buf, false );
    }


BitIOStreamBuffer BitOStream::reset()
    {
    return unbind();
    }

    
BitIOStreamBuffer BitOStream::get_buffer()
    {
    return buffer;
    }


void BitOStream::align_to_byte()
    {
    if( bit > 0 )
        {
        write_int( 0, 8 - bit );
        hope( !bit );
        }
    }


void BitOStream::write_int
    (
    UInt32  value,
    int     bits
    )
    {
    hope( 0 < bits );
    hope( bits <= 32 );

    // Get rid of high bits.

    if( bits < 32 )
        {
        value &= ( 1 << bits ) - 1;
        }

    // Write unaligned bits.

    int shift;
    if( bit )
        {
        BitIOStreamBuffer::Iterator last( buffer.end() - 1 );
        (*last) |= static_cast< UInt8 >( value << bit );
        shift = 8 - bit;
        }
    else
        {
        buffer.push_back( static_cast< UInt8 >( value ) );
        shift = 8;
        }

    // Update bit pointer and counter.

    bit = ( bit + bits ) & 7;
    num_bits += bits;
    bits -= shift;

    // Write aligned bytes.

    if( bits > 0 )
        {
        value >>= shift;
        do
            {
            buffer.push_back( static_cast< UInt8 >( value ) );
            value >>= 8;
            bits -= 8;
            }
        while( bits > 0 );
        }
    }


void BitOStream::write_int64
    (
    UInt64  value,
    int     bits
    )
    {
    hope( 0 < bits );
    hope( bits <= 64 );

    // Get rid of high bits.

    if( bits < 64 )
        {
        value &= ( UINT64( 1 ) << bits ) - 1;
        }

    // Write unaligned bits.

    int shift;
    if( bit )
        {
        BitIOStreamBuffer::Iterator last( buffer.end() - 1 );
        (*last) |= static_cast< UInt8 >( value << bit );
        shift = 8 - bit;
        }
    else
        {
        buffer.push_back( static_cast< UInt8 >( value ) );
        shift = 8;
        }

    // Update bit pointer and counter.

    bit = ( bit + bits ) & 7;
    num_bits += bits;
    bits -= shift;

    // Write aligned bytes.

    if( bits > 0 )
        {
        value >>= shift;
        do
            {
            buffer.push_back( static_cast< UInt8 >( value ) );
            value >>= 8;
            bits -= 8;
            }
        while( bits > 0 );
        }
    }


void BitOStream::write_unary
    (
    UInt32 value
    )
    {
    hope( value >= 0 && value <= 64 );
    if( value > 0 )
        {
        // Write @a value zeroes.

        write_int64( 0, value );
        }
    write_int( 1, 1 );
    }


void BitOStream::write_float
    (
    Float32 value
    )
    {
    write_int( *reinterpret_cast<UInt32 *>( &value ), 32 );
    }


void BitOStream::write_float64
    (
    Float64 value
    )
    {
    write_int64( *reinterpret_cast<UInt64 *>( &value ), 64 );
    }


int BitOStream::bit_size() const
    {
    return num_bits;
    }


int BitOStream::byte_size() const
    {
    return buffer.size();
    }


} //namespace Massiv::Core
} //namespace Massiv
